@@ -28,8 +28,8 @@ import textwrap
 import time
 import StringIO
 import requests
+from copy import deepcopy
 
-__version__ = '1.0a'
 
 """
 This is my uniprot python library. It provides a bunch of 
@@ -46,10 +46,6 @@ including the protein sequence.
 """
 
 
-# Doing internet lookups is fraught with error, 
-# so caching to disk is recommended. Here's a
-# couple of helper functions to cache Python dictionaries.
-
 
 def is_html(text):
   if re.search('<html', text):
@@ -107,7 +103,7 @@ def batch_uniprot_id_mapping_pairs(
   while i_seqid <= len(seqids):
     seqids_subset = seqids[i_seqid:i_seqid+batch_size]
     if cache_basename:
-      subset_cache = '%s%d.txt' % (cache_basename, i_seqid)
+      subset_cache = '%s.%d.txt' % (cache_basename, i_seqid)
     else:
       subset_cache = None
     subset_pairs = get_uniprot_id_mapping_pairs(
@@ -117,6 +113,103 @@ def batch_uniprot_id_mapping_pairs(
   return pairs
 
 
+def parse_isoforms(text):
+  """
+  Returns a dictionary of uniprot_acc and entries.
+
+  Each entry contains:
+    - var_seq: the sequence variations in the file
+    - isoforms: the interpreted isoform sequences with
+                the seqids
+  """
+  tag = None
+  uniprot_data = {}
+  var_seq = None
+  in_isoform_section = False
+  isoform_id = None
+  for l in text.splitlines():
+    test_tag = l[:5].strip()
+    if test_tag and test_tag != tag:
+      tag = test_tag
+    line = l[5:].strip()
+    words = line.split()
+    if tag == "ID":
+      uniprot_id = words[0]
+      uniprot_data[uniprot_id] = {
+        'var_seqs': [],
+        'isoforms': {},
+        'sequence': '',
+      }
+    if tag == "SQ":
+      if words[0] != "SEQUENCE":
+        uniprot_data[uniprot_id]['sequence'] += ''.join(words)
+    if tag == 'FT':
+      if var_seq is not None and l[5] != ' ':
+        var_seq = None
+      if line.startswith('VAR_SEQ'):
+        var_seq = {
+          'i': int(words[1]),
+          'j': int(words[2]),
+          'block': ''
+        }
+        uniprot_data[uniprot_id]['var_seqs'].append(var_seq)
+      if var_seq is not None:
+        var_seq['block'] += l[34:]
+        if l.endswith('isoform'):
+          # needed later to search isoform references
+          var_seq['block'] += ' '
+    if tag == 'CC':
+      if words[0] == '-!-':
+        if 'ALTERNATIVE' in words[1] and 'PRODUCTS' in words[2]:
+          in_isoform_section = True
+        else:
+          in_isoform_section = False
+      if in_isoform_section:
+        if words[0].startswith('Name'):
+          isoform_id = str(words[0][:-1].split('=')[1])
+        for word in words:
+          if word.startswith('IsoId='):
+            seqid = word[:-1].split('=')[1]
+            uniprot_data[uniprot_id]['isoforms'][isoform_id] = {
+              'seqid': seqid
+            }
+            break
+  for uniprot_id in uniprot_data:
+    var_seqs = uniprot_data[uniprot_id]['var_seqs']
+    isoforms = uniprot_data[uniprot_id]['isoforms']
+    original_sequence = uniprot_data[uniprot_id]['sequence']
+    for var_seq in var_seqs:
+      block = var_seq['block']
+      match = re.search(r'\(.*\)', block)
+      isoform_tokens = match.group()[1:-1].split()
+      isoform_ids = []
+      for i in range(len(isoform_tokens)):
+        if 'isoform' in isoform_tokens[i]:
+          isoform_ids.append(str(isoform_tokens[i+1]))
+      var_seq['isoform_ids'] = isoform_ids
+      if block.startswith('Missing'):
+        var_seq['deletion'] = True
+      else:
+        var_seq['deletion'] = False
+        transition = block.split('(')[0]
+        original, mutation = transition.split('->')
+        var_seq['sequence'] = original.strip()
+        assert len(var_seq['sequence']) == var_seq['j'] - var_seq['i'] + 1
+        var_seq['mutation'] = mutation.strip()
+    var_seqs.sort(key=lambda v:-v['i'])
+    for isoform_id in isoforms:
+      sequence = original_sequence
+      for var_seq in uniprot_data[uniprot_id]['var_seqs']:
+        if isoform_id in var_seq['isoform_ids']:
+          i = var_seq['i']-1
+          j = var_seq['j']
+          if var_seq['deletion']:
+            sequence = sequence[:i] + sequence[j:]
+          else:
+            sequence = sequence[:i] + var_seq['mutation'] + sequence[j:]
+      isoforms[isoform_id]['sequence'] = sequence      
+  return uniprot_data
+
 
 def parse_uniprot_txt_file(cache_txt):
   """
@@ -133,7 +226,7 @@ def parse_uniprot_txt_file(cache_txt):
 
   tag = None
   uniprot_id = None
-  uniprot_data = {}
+  metadata_by_seqid = {}
   for l in cache_txt.splitlines():
     test_tag = l[:5].strip()
     if test_tag and test_tag != tag:
@@ -144,14 +237,14 @@ def parse_uniprot_txt_file(cache_txt):
       uniprot_id = words[0]
       is_reviewed = words[1].startswith('Reviewed')
       length = int(words[2])
-      uniprot_data[uniprot_id] = {
+      metadata_by_seqid[uniprot_id] = {
         'id': uniprot_id,
         'is_reviewed': is_reviewed,
         'length': length,
         'sequence': '',
         'accs': [],
       }
-      entry = uniprot_data[uniprot_id]
+      entry = metadata_by_seqid[uniprot_id]
     if tag == "SQ":
       if words[0] != "SEQUENCE":
         entry['sequence'] += ''.join(words)
@@ -202,7 +295,7 @@ def parse_uniprot_txt_file(cache_txt):
         entry['comment'] = ''
       entry['comment'] += line + '\n'
 
-  for entry in uniprot_data.values():
+  for entry in metadata_by_seqid.values():
     descriptions = entry['descriptions']
     for i in reversed(range(len(descriptions))):
       description = descriptions[i]
@@ -213,9 +306,38 @@ def parse_uniprot_txt_file(cache_txt):
         del descriptions[i]
     entry['description'] = '; '.join(descriptions)
 
-  return uniprot_data
+  return metadata_by_seqid
 
 
+def parse_uniprot_metadata_with_seqids(seqids, cache_txt):
+  """
+  Returns a dictionary of metadata of given seqids, doing the requisite
+  lookup of seqids in the ACCs and IDs, and more importantly handles
+  isoform seqid's
+  """
+  metadata = parse_uniprot_txt_file(cache_txt)
+  for uniprot_id in metadata.keys():
+    for seqid in metadata[uniprot_id]['accs'] + [metadata[uniprot_id]['id']]:
+      metadata[seqid] = metadata[uniprot_id]
+  results = {}
+  isoform_dict = parse_isoforms(cache_txt)
+  for seqid in seqids:
+    if seqid in metadata:
+      results[seqid] = metadata[seqid]
+    else:
+      primary_seqid = seqid[:6]
+      if primary_seqid in metadata:
+        protein_metadata = metadata[primary_seqid]
+        uniprot_id = protein_metadata['id']
+        if uniprot_id in isoform_dict:
+          isoforms = isoform_dict[uniprot_id]['isoforms'].values()
+          for isoform in isoforms:
+            if isoform['seqid'] == seqid:
+              results[seqid] = deepcopy(protein_metadata)
+              results[seqid]['accs'] = [seqid]
+              results[seqid]['sequence'] = isoform['sequence']
+  return results
+
 
 def fetch_uniprot_metadata(seqids, cache_fname=None):
   """
@@ -223,17 +345,18 @@ def fetch_uniprot_metadata(seqids, cache_fname=None):
   by parse_uniprot_txt_file) of the given seqids. The seqids
   must be valid uniprot identifiers.
 
-  uniprot can't handle isoforms.
+  Now handles isoform versions of accession id's!
   """
 
+  primary_seqids = [s[:6] for s in seqids]
   if cache_fname and os.path.isfile(cache_fname):
     print "Loading uniprot metadata from", cache_fname
     cache_txt = open(cache_fname).read()
   else:
-    print "Fetching uniprot metadata for %d ACC seqids ..." % len(seqids)
+    print "Fetching uniprot metadata for %d ACC seqids ..." % len(primary_seqids)
     r = requests.post(
         'http://www.uniprot.org/batch/', 
-        files={'file':StringIO.StringIO(' '.join(seqids))}, 
+        files={'file':StringIO.StringIO(' '.join(primary_seqids))}, 
         params={'format':'txt'})
     while 'Retry-After' in r.headers:
       t = int(r.headers['Retry-After'])
@@ -248,117 +371,7 @@ def fetch_uniprot_metadata(seqids, cache_fname=None):
       print "Error in fetching metadata"
       return {}
 
-  metadata = parse_uniprot_txt_file(cache_txt)
-
-  # resort the dictionary wrt to input seqids as keys
-  results = {}
-  for uniprot_id in metadata.keys():
-    for seqid in metadata[uniprot_id]['accs'] + [metadata[uniprot_id]['id']]:
-      if seqid in seqids:
-        results[seqid] = metadata[uniprot_id]
-  for seqid in seqids:
-    if seqid not in results:
-      primary_seqid = seqid[:6]
-      if primary_seqid in results:
-        results[seqid] = results[primary_seqid]
-  return results
-
-
-
-def parse_isoforms(text):
-  """
-  Returns a dictionary of uniprot_acc and entries.
-
-  Each entry contains:
-    - var_seq: the sequence variations in the file
-    - isoforms: the interpreted isoform sequences with
-                the seqids
-  """
-  tag = None
-  uniprot_data = {}
-  var_seq = None
-  in_isoform_section = False
-  isoform_id = None
-  for l in text.splitlines():
-    test_tag = l[:5].strip()
-    if test_tag and test_tag != tag:
-      tag = test_tag
-    line = l[5:].strip()
-    words = line.split()
-    if tag == "ID":
-      uniprot_id = words[0]
-      uniprot_data[uniprot_id] = {
-        'var_seqs': [],
-        'isoforms': {},
-        'sequence': '',
-      }
-    if tag == "SQ":
-      if words[0] != "SEQUENCE":
-        uniprot_data[uniprot_id]['sequence'] += ''.join(words)
-    if tag == 'FT':
-      if var_seq is not None and l[5] != ' ':
-        var_seq = None
-      if line.startswith('VAR_SEQ'):
-        var_seq = {
-          'i': int(words[1]),
-          'j': int(words[2]),
-          'block': ''
-        }
-        uniprot_data[uniprot_id]['var_seqs'].append(var_seq)
-      if var_seq is not None:
-        var_seq['block'] += l[34:]
-    if tag == 'CC':
-      if words[0] == '-!-':
-        if 'ALTERNATIVE' in words[1] and 'PRODUCTS' in words[2]:
-          in_isoform_section = True
-        else:
-          in_isoform_section = False
-      if in_isoform_section:
-        if words[0].startswith('Name'):
-          isoform_id = int(words[0][:-1].split('=')[1])
-        for word in words:
-          if word.startswith('IsoId='):
-            seqid = word[:-1].split('=')[1]
-            uniprot_data[uniprot_id]['isoforms'][isoform_id] = {
-              'seqid': seqid
-            }
-            break
-  for uniprot_id in uniprot_data:
-    var_seqs = uniprot_data[uniprot_id]['var_seqs']
-    isoforms = uniprot_data[uniprot_id]['isoforms']
-    original_sequence = uniprot_data[uniprot_id]['sequence']
-    for var_seq in var_seqs:
-      block = var_seq['block']
-      match = re.search(r'\(.*\)', block)
-      isoform_tokens = match.group()[1:-1].split()
-      isoform_ids = []
-      for i in range(len(isoform_tokens)):
-        if 'isoform' in isoform_tokens[i]:
-          isoform_ids.append(int(isoform_tokens[i+1]))
-      var_seq['isoform_ids'] = isoform_ids
-      if block.startswith('Missing'):
-        var_seq['deletion'] = True
-      else:
-        var_seq['deletion'] = False
-        transition = block.split('(')[0]
-        original, mutation = transition.split('->')
-        var_seq['sequence'] = original.strip()
-        assert len(var_seq['sequence']) == var_seq['j'] - var_seq['i'] + 1
-        var_seq['mutation'] = mutation.strip()
-    var_seqs.sort(key=lambda v:-v['i'])
-    for isoform_id in isoforms:
-      sequence = original_sequence
-      for var_seq in uniprot_data[uniprot_id]['var_seqs']:
-        if isoform_id in var_seq['isoform_ids']:
-          i = var_seq['i']-1
-          j = var_seq['j']
-          if var_seq['deletion']:
-            sequence = sequence[:i] + sequence[j:]
-          else:
-            sequence = sequence[:i] + var_seq['mutation'] + sequence[j:]
-      isoforms[isoform_id]['sequence'] = sequence      
-  return uniprot_data
-
+  return parse_uniprot_metadata_with_seqids(seqids, cache_txt)
 
 
 def batch_uniprot_metadata(seqids, cache_basename=None, batch_size=500):
@@ -367,7 +380,7 @@ def batch_uniprot_metadata(seqids, cache_basename=None, batch_size=500):
   by parse_uniprot_txt_file) of the given seqids. The seqids
   must be valid uniprot identifiers.
 
-  uniprot can't handle isoforms.
+  Now handles isoform versions of accession id's!
   """
 
   unique_seqids = list(set(seqids))
@@ -378,7 +391,7 @@ def batch_uniprot_metadata(seqids, cache_basename=None, batch_size=500):
   while i_seqid <= len(unique_seqids):
     seqids_subset = unique_seqids[i_seqid:i_seqid+batch_size]
     if cache_basename:
-      subset_cache = '%s%d.txt' % (cache_basename, i_seqid)
+      subset_cache = '%s.%d.txt' % (cache_basename, i_seqid)
     else:
       subset_cache = None
     metadata_subset = fetch_uniprot_metadata(
@@ -471,7 +484,7 @@ def probe_id_type(entries, is_id_fn, name, uniprot_mapping_type, cache_fname):
   alternative_to_uniprot = { p[0]:p[1] for p in pairs }
   for entry in entries:
     if entry['seqid'] in alternative_to_uniprot:
-      entry['uniprot_id'] = alternative_to_uniprot[entry['seqid']]
+      entry['uniprot_acc'] = alternative_to_uniprot[entry['seqid']]
 
 
 def get_metadata_with_some_seqid_conversions(seqids, cache_fname=None):
@@ -482,7 +495,7 @@ def get_metadata_with_some_seqid_conversions(seqids, cache_fname=None):
       'raw_seqid':seqid,
       'seqid':'',
       'id_type':'',
-      'uniprot_id':'',
+      'uniprot_acc':'',
       'metadata':''})
   
   # break up pieces when in form xx|xxxxx|xxx
@@ -491,34 +504,50 @@ def get_metadata_with_some_seqid_conversions(seqids, cache_fname=None):
 
   # convert a few types into uniprot_ids
   id_types = [
-    (is_sgd, 'ordered-locus-tag', 'ENSEMBLGENOME_PRO_ID'),
-    (is_refseq, 'refseq', 'P_REFSEQ_AC'),
-    (is_refseq, 'refseq', 'REFSEQ_NT_ID'),
+    (is_sgd, 'locustag', 'ENSEMBLGENOME_PRO_ID'),
+    (is_refseq, 'refseqp', 'P_REFSEQ_AC'),
+    (is_refseq, 'refseqnt', 'REFSEQ_NT_ID'),
     (is_ensembl, 'ensembl', 'ENSEMBL_ID'),
-    (is_maybe_uniprot_id, 'uniprot-entity', 'ID')]
+    (is_maybe_uniprot_id, 'uniprotid', 'ID')]
   for is_id_fn, name, uniprot_mapping_type in id_types:
-    probe_id_type(entries, is_id_fn, name, uniprot_mapping_type, cache_fname+'.'+name)
+    if cache_fname:
+      seqid_cache_fname = cache_fname+'.'+name
+    else:
+      seqid_cache_fname = None
+    probe_id_type(entries, is_id_fn, name, uniprot_mapping_type, seqid_cache_fname)
 
-  # delete the variant suffixes in some uniprot id's
+  # delete the variant suffixes in some uniprot id's as probe_id_type
+  # can't cope with uniprot variant for id mapping lookup (bad!)
   for entry in entries:
     if entry['id_type'] == '' and is_uniprot_variant(entry['seqid']):
       entry['seqid'] = entry['seqid'][:6]
 
   # map UNIPROT ID's to their current best entry
-  probe_id_type(entries, is_uniprot, 'UNIPROT-ACC', 'ACC+ID', cache_fname+'.seqid')
+  if cache_fname:
+    seqid_cache_fname = cache_fname+'.seqid'
+  else:
+    seqid_cache_fname = None
+  probe_id_type(entries, is_uniprot, 'UNIPROT-ACC', 'ACC+ID', seqid_cache_fname)
 
-  uniprot_seqids = [entry['uniprot_id'] for entry in entries 
-                    if 'uniprot_id' in entry]
+  uniprot_seqids = []
+  for entry in entries:
+    if 'uniprot_acc' in entry:
+      uniprot_seqids.append(entry['uniprot_acc'])
+    if is_uniprot_variant(entry['raw_seqid']):
+      # put the isoform variants back up id name as
+      # batch_uniprot_metadata can handle isoforms!
+      uniprot_seqids.append(entry['raw_seqid'])
+      entry['uniprot_acc'] = entry['raw_seqid']
   uniprot_dict = batch_uniprot_metadata(uniprot_seqids, cache_fname)
 
   result = {}
   for entry in entries:
-    if entry['uniprot_id'] == "":
+    if entry['uniprot_acc'] == "":
       continue
-    uniprot_id = entry['uniprot_id']
-    if uniprot_id not in uniprot_dict:
+    uniprot_acc = entry['uniprot_acc']
+    if uniprot_acc not in uniprot_dict:
       continue
-    result[entry['raw_seqid']] = uniprot_dict[uniprot_id]
+    result[entry['raw_seqid']] = uniprot_dict[uniprot_acc]
 
   return result
 
